/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2002/02/25
 *
 * $Id$
 */


/*

RFC1890

4.4.6 GSM

   GSM (group speciale mobile) denotes the European GSM 06.10
   provisional standard for full-rate speech transcoding, prI-ETS 300
   036, which is based on RPE/LTP (residual pulse excitation/long term
   prediction) coding at a rate of 13 kb/s [7,8,9]. The standard can be
   obtained from

*/

#include <sys/param.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include "aucodec.h"
#include "gsm.h"
#include "debug.h"

#define ENDIAN_LITTLE


#if 0
typedef enum{
	AUCODEC_DSPFMT_U8    = 1<<0,
	AUCODEC_DSPFMT_S16LE = 1<<1,
	AUCODEC_DSPFMT_S16BE = 1<<2,
	AUCODEC_DSPFMT_S8    = 1<<3,
	AUCODEC_DSPFMT_U16LE = 1<<4,
	AUCODEC_DSPFMT_U16BE = 1<<5,
}aucodec_dspfmt_t;
#endif

typedef struct _aucodec_gsm_enc_t {
	aucodec_enc_t        aucodec_enc;
	gsm                  gsm_handle;
}aucodec_gsm_enc_t;

typedef struct _aucodec_gsm_dec_t {
	aucodec_dec_t        aucodec_dec;
	gsm                  gsm_handle;
}aucodec_gsm_dec_t;


static void      aucodec_gsm_enc
			(aucodec_enc_t        *enc,
			 void                 *src,
			 void                 *dst,
			 u_int32_t            *ret_dst_len);
static void      aucodec_gsm_enc_deinit
                        (aucodec_enc_t        *enc);
static void      aucodec_gsm_dec
			(aucodec_dec_t        *dec,
			 void                 *src,
			 u_int32_t             src_len,
			 void                 *dst);
static void      aucodec_gsm_dec_deinit
                        (aucodec_dec_t        *dec);

u_int32_t aucodec_gsm_enc_init
			(aucodec_dspfmt_t   dspfmt_avail, 
			 u_int32_t          ch,
			 u_int32_t          pcmblk_len, 
			 aucodec_dspfmt_t  *ret_dspfmt,
			 aucodec_enc_t    **ret_enc){
	aucodec_gsm_enc_t *enc;
	int ret;

	enc=malloc(sizeof(aucodec_gsm_enc_t));
	if(enc==NULL){
		return 1;
	}
	*ret_enc=(aucodec_enc_t*)enc;
	ret=2;
	if(dspfmt_avail & AUCODEC_DSPFMT_S16HE &&
	   pcmblk_len==160 &&
	   ch==1){
		enc->gsm_handle=gsm_create();
		if(enc->gsm_handle==NULL){
			ret=3;goto free1;
		}
		enc->aucodec_enc.enc=aucodec_gsm_enc;
		enc->aucodec_enc.deinit=aucodec_gsm_enc_deinit;
		enc->aucodec_enc.codeblk_maxsize=33;
		enc->aucodec_enc.pcmblk_size=160*2;
		*ret_dspfmt=AUCODEC_DSPFMT_S16HE;
		return 0;
	}
free1:
	free(enc);
	return ret;
}

static void      aucodec_gsm_enc
			(aucodec_enc_t        *enc,
			 void                 *src,
			 void                 *dst,
			 u_int32_t            *ret_dst_len){

	aucodec_gsm_enc_t *gsmenc;

	gsmenc=(aucodec_gsm_enc_t*)enc;
	gsm_encode(gsmenc->gsm_handle,(gsm_signal*)src,(gsm_byte*)dst);
	*ret_dst_len=33;
	return;
}

static void      aucodec_gsm_enc_deinit
                        (aucodec_enc_t        *enc){
	aucodec_gsm_enc_t *gsmenc;

	gsmenc=(aucodec_gsm_enc_t*)enc;
	gsm_destroy(gsmenc->gsm_handle);
	free(gsmenc);
	return;
}

u_int32_t aucodec_gsm_dec_init
			(aucodec_dspfmt_t   dspfmt_avail,
			 u_int32_t          ch,
			 u_int32_t          pcmblk_len,
			 aucodec_dspfmt_t  *ret_dspfmt,
			 aucodec_dec_t    **ret_dec){
	aucodec_gsm_dec_t *dec;
	int ret;

	dec=malloc(sizeof(aucodec_gsm_dec_t));
	if(dec==NULL){
		return 1;
	}
	*ret_dec=(aucodec_dec_t*)dec;
	ret=2;
	if(dspfmt_avail & AUCODEC_DSPFMT_S16HE &&
	   pcmblk_len==160 &&
	   ch==1){
		dec->gsm_handle=gsm_create();
		if(dec->gsm_handle==NULL){
			ret=3;goto free1;
		}
		dec->aucodec_dec.dec=aucodec_gsm_dec;
		dec->aucodec_dec.deinit=aucodec_gsm_dec_deinit;
		dec->aucodec_dec.codeblk_maxsize=33;
		dec->aucodec_dec.pcmblk_size=160*2;
		*ret_dspfmt=AUCODEC_DSPFMT_S16HE;
		return 0;
	}
free1:
	free(dec);
	return ret;
}

static void      aucodec_gsm_dec
			(aucodec_dec_t        *dec,
			 void                 *src,
			 u_int32_t             src_len,
			 void                 *dst){
	aucodec_gsm_dec_t *gsmdec;
	int gsm_ret;
	
	gsmdec=(aucodec_gsm_dec_t*)dec;
	gsm_ret=gsm_decode(gsmdec->gsm_handle,(gsm_byte*)src,(gsm_signal*)dst);
	return;
}
static void      aucodec_gsm_dec_deinit
                        (aucodec_dec_t        *dec){
	aucodec_gsm_dec_t *gsmdec;
	
	gsmdec=(aucodec_gsm_dec_t*)dec;
	gsm_destroy(gsmdec->gsm_handle);
	free(gsmdec);
	return;
}


/* EOF */
