/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2001/0?/??
 *
 * $Id: audioplay.c,v 1.5 2001/10/19 01:13:21 kasamatu Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <machine/soundcard.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#include "aucodec.h"
#include "debug.h"
#include "pipe_hd.h"
#include "def_value.h"
#include "timeval.h"

#ifdef DEBUG
#define debug(format, args...)\
        fprintf(stderr,"%s:%d:"format,__PRETTY_FUNCTION__, __LINE__ , ##args);
#else
#define debug(format, args...)\
{}
#endif


#define ENDIAN_LITTLE

/*0x0000ffff log(len)  [0=>E(0=>1<<E) es1371_OSS]*/
/*0xffff0000 # of frag (is ignored.)*/
#define DEFAULT_DSP_FRAGMENT_ARG  (0xff0008) 

#define DEFAULT_DSP_DEVICE      "/dev/dsp"
#define DEFAULT_DSP_FD          (-1)

#define DEFAULT_AUDIOPLAY_DECODER_INIT (aucodec_l16_dec_init)
#define DEFAULT_AUDIOPLAY_RATE         (44100)
#define DEFAULT_AUDIOPLAY_CHANNEL      (2)

#define DEFAULT_EMON_DELAY      (OPT.emon_freq+1)
#define DEFAULT_EMONAUBUF_NUM   (3)

//#define DSP_RATE_BASE (760)/*ESS Maestoro OSS / Toshiba DynaBookSS3380V*/
#define DSP_RATE_BASE (0)

typedef struct _audioplay_arg_t {
	char              *dsp_dev;
	int                emon_clk; 
	int                emon_freq;
        int                plen;
	int                dsp_rate;
	int                dsp_ch;
        aucodec_dec_init_f decoder_init;
        int32_t          constdelay_ts;

	int                dsp_fd;  /*if not==-1,dsp opened */
	int                dsp_fmt;
} audioplay_arg_t;

static audioplay_arg_t     OPT;
static aucodec_dec_t      *decoder;

u_int32_t audioplay_parsearg(int argc, char *argv[],audioplay_arg_t *opt);
int             open_dsp(void);

static void audioplay_db_printobuf(int fd);

int
main(int argc, char **argv)
{
	unsigned int    timestamp;
	u_int8_t        p_header[PIPE_HEADER_LEN];

	void           *pcmbuf;
	void           *codebuf;
	int             afd;	/* audio fd */
	u_int64_t       dclk_tsnow;      /* dsp clock */
	u_int32_t       wclk_tsstart;
	struct timeval  wclk_ps_tvstart; /* the timeval before start capture */
	struct timeval  wclk_tvstart, wclk_tvnow;
	ssize_t         io_ret; /* return value for read write seek ..*/ 

	u_int64_t       frame_cnt;
	u_int32_t       dsp_ossbuflen;
	if (audioplay_parsearg(argc, argv,&OPT) > 0) {
		return -1;
	}
	d1_printf(
		"emon clock %d freq %d plen %d\n"
		"dsp rate %d dsp channel %d\n"
		,OPT.emon_clk, OPT.emon_freq,OPT.plen
		,OPT.dsp_rate, OPT.dsp_ch);
        if (((OPT.dsp_rate*OPT.emon_freq) % OPT.emon_clk) != 0){
	  e_printf("Samples per FEC block"
		   " (=sampling_rate*(emon_freq/emon_clock))"
		   " must be integar.\n");
                return -1;
        }
	if ((afd = open_dsp()) < 0) {
		return -1;
	}
	d1_printf(
		"fmt=%s dspfmt 0x%x dspbuf %u decbuf %u\n"
		,aucodec_dec_init2str(OPT.decoder_init),OPT.dsp_fmt
		,decoder->pcmblk_size,decoder->codeblk_maxsize);


//	afd=OPT.dsp_fd;
	pcmbuf = malloc(decoder->pcmblk_size);
	codebuf= malloc(decoder->codeblk_maxsize);
	if (pcmbuf ==NULL || codebuf == NULL) {
		e_printf("fail to  malloc %d bytes\n"
			 ,decoder->pcmblk_size+decoder->codeblk_maxsize);
		return -1;
	}
#if 0
	{
		audio_buf_info aubuf_info;
		/* if read dsp too late , drop samples. */
		ioctl(afd, SNDCTL_DSP_GETOSPACE, &aubuf_info);
		/* todo : should be based on OPT.delay_limit */
		dsp_ossbuflen=aubuf_info.bytes;
		d1_printf("audioplay startup.auinfo.bytes=%d\n",dsp_ossbuflen);
	}
#endif

	gettimeofday(&wclk_ps_tvstart, NULL);
#if 0
	audioplay_db_printobuf(afd);
	sleep(1);
#endif
	for (;;) {
		int32_t   read_len;
		d3_printf("read start\n",read_len);
		read_len = pipe_blocked_read_block(STDIN_FILENO, p_header
						   ,codebuf
						   ,decoder->codeblk_maxsize);
		d3_printf("read %dbytes\n",read_len);
		if (read_len < 0) {
			e_printf("fail to read sample.\n");
			continue;
		}
		if (read_len == 0) {
			e_printf("packet loss. ts=%d\n",
				 pipe_get_timestamp(p_header));
			continue;
		}
		decoder->dec(decoder,codebuf,read_len,pcmbuf);
		{int32_t write_len;
		write_len = write(afd, pcmbuf,decoder->pcmblk_size);
//		audioplay_db_printobuf(afd);
		if(write_len!=decoder->pcmblk_size){
			e_printf("dsp obuf full?? %d=>%d\n"
				 ,decoder->pcmblk_size,write_len);
//			audioplay_db_printobuf(afd);
		}
		}
	}
	d_printf("\n");
	return 0;
}



int
open_dsp(void)
{
	int             afd;
	int             val;
	u_int32_t       ret;
	aucodec_dspfmt_t dspfmt_avail,dspfmt;

	/*audiocaptplay*/
	if (OPT.dsp_fd!=-1){
		dspfmt_avail=aucodec_dspfmt_oss2dspfmt(OPT.dsp_fmt);
		ret=OPT.decoder_init(dspfmt_avail,
				 OPT.dsp_ch,
				 (OPT.dsp_rate*OPT.emon_freq)/OPT.emon_clk,
				 &dspfmt,
				 &decoder);
		if(ret!=0){
			e_printf("fail to initialize decoder.\n");
			return -1;
		}
		OPT.dsp_fmt=aucodec_dspfmt_dspfmt2oss(dspfmt);
#if 0
		return OPT.dsp_fd;
#else
		return STDOUT_FILENO;
#endif
	}
	
	ret=OPT.decoder_init(AUCODEC_DSPFMT_S16LE,
			 OPT.dsp_ch,
			 (OPT.dsp_rate*OPT.emon_freq)/OPT.emon_clk,
			 &dspfmt,
			 &decoder);
	return STDOUT_FILENO;
	/* open audio device */
	if ((afd = open(OPT.dsp_dev, O_WRONLY, 0)) < 0) {
		e_printf("cannot open audio device\n");
		return -1;
	}
	OPT.dsp_fd = afd;

#if 1
	val = DEFAULT_DSP_FRAGMENT_ARG;
	if (ioctl(afd, SNDCTL_DSP_SETFRAGMENT, &val) < 0) {
		e_printf("SNDCTL_DSP_SETFRAGMENT: cannot set audio device \n");
		return -1;
	}
#else
	val = OPT.fragment_size;/* fragment size  (is not for DMA ...) */
	if (ioctl(afd, SNDCTL_DSP_SETBLKSIZE, &val) < 0) {
		e_printf("SNDCTL_DSP_SETBLKSIZE: cannot set audio device \n");
		return -1;
	}
#endif

	/* get supported format. */
	val = AFMT_QUERY;
	if (ioctl(afd, SNDCTL_DSP_GETFMTS, &val) < 0) {
		e_printf("SNDCTL_DSP_GETFMTS: cannot get format mask.\n");
		return -1;
	}
	d1_printf("SNDCTL_DSP_GETFMTS ret 0x%x\n", val);

	dspfmt_avail=aucodec_dspfmt_oss2dspfmt(val);
	ret=OPT.decoder_init(dspfmt_avail,
			 OPT.dsp_ch,
			 (OPT.dsp_rate*OPT.emon_freq)/OPT.emon_clk,
			 &dspfmt,
			 &decoder);
	if(ret!=0){
		e_printf("fail to initialize decoder.\n");
		return -1;
	}

	/* set audio device */
	val = OPT.dsp_ch;
	/* old FreeBSD don't support SNDCTL_DSP_CHANNELS  */
	if (ioctl(afd, SOUND_PCM_WRITE_CHANNELS, &val) < 0) {
		printf("SNDCTL_DSP_CHANNELS: cannot set channel\n");
		return -1;
	}

	OPT.dsp_fmt=aucodec_dspfmt_dspfmt2oss(dspfmt);
	val=OPT.dsp_fmt;
	d1_printf("SNDCTL_DSP_SETFMT 0x%x\n", val);
	if (ioctl(afd, SNDCTL_DSP_SETFMT, &val) < 0) {
		e_printf("SNDCTL_DSP_SETFMT: cannot set format. \n");
		return -1;
	}

	OPT.dsp_rate+=DSP_RATE_BASE;
	val = OPT.dsp_rate;
	if (ioctl(afd, SNDCTL_DSP_SPEED, &val) < 0) {
		e_printf("SNDCTL_DSP_SPEED: cannot set dsp speed.\n");
		return -1;
	}
	if (val !=OPT.dsp_rate) {
		printf("cannot set SNDCTL_DSP_SPEED to %d\n", OPT.dsp_rate);
		return -1;
	}
	return afd;
}


u_int32_t audioplay_parsearg(int argc, char *argv[],audioplay_arg_t *opt){
        int32_t         constdelay;
	char            c;
	char           *tmpstr;
	char           *opts = "d:C:R:L:f:r:c:D:X:Y:"; /*X,Y for recplay*/
	char           *helpmes =
	"usage : %s [-C <clock>] [-R <freq>] [-f <au_fmt>] "
	"[-r <dsp_rate>] [-c <chanel>]\n"
	" options\n"
//	"  d <s>   : dsp path\n"
	"  C <n>   : EMON clock\n"
	"  R <n>   : EMON frequency\n"
	"  L <n>   : emon payload length limit\n"
	"  f <s>   : audio format (L8,L16)\n"
	"  r <n>   : sampling rate\n"
	"  c <n>   : # of chanel\n"
	"  d <n>   : constant delay # msec\n"
	" debug option\n"
	"  D <n>   : debug message level\n"
		;
	opt->dsp_dev   = DEFAULT_DSP_DEVICE;
	opt->dsp_fd    = DEFAULT_DSP_FD;
	
	opt->emon_clk  = DEF_EMON_CLOCK;
	opt->emon_freq = DEF_EMON_FREQ;
	opt->plen      = DEF_EMON_PLEN;

	opt->dsp_rate  = DEFAULT_AUDIOPLAY_RATE;
	opt->dsp_ch    = DEFAULT_AUDIOPLAY_CHANNEL;
	opt->decoder_init   = DEFAULT_AUDIOPLAY_DECODER_INIT;

	constdelay=-1;

	if ((tmpstr = getenv("EMON_CLOCK")) != NULL)
		opt->emon_clk = atoi(tmpstr);
	if ((tmpstr = getenv("EMON_FREQ")) != NULL)
		opt->emon_freq = atoi(tmpstr);
	if ((tmpstr = getenv("EMON_PLEN")) != NULL)
		opt->plen = atoi(tmpstr);

	if ((tmpstr = getenv("AUDIOPLAY_RATE")) != NULL)
		opt->dsp_rate = atoi(tmpstr);
	if ((tmpstr = getenv("AUDIOPLAY_CHANNEL")) != NULL)
		opt->dsp_ch = atoi(tmpstr);
	if ((tmpstr = getenv("AUDIOPLAY_FORMAT")) != NULL) {
		opt->decoder_init=aucodec_dec_str2init(tmpstr);
		if(opt->decoder_init==NULL){
			e_printf("AUDIOPLAY_FORMAT=%s is not supported.\n"
			       , tmpstr);
			return 2;
		}
	}

	while ((c = getopt(argc, argv, opts)) != -1) {
		switch (c) {
/* 		case 'd': */
/* 			opt->dsp_dev=optarg; */
/* 			break; */
		case 'C':
			opt->emon_clk = atoi(optarg);
			break;
		case 'R':
			opt->emon_freq = atoi(optarg);
			break;
		case 'L':
			opt->plen = atoi(optarg);
			break;
		case 'f':
			opt->decoder_init=aucodec_dec_str2init(optarg);
			if(opt->decoder_init==NULL){
				printf("AUDIOPLAY_FORMAT %s is not supported.\n"
				       , optarg);
				return 2;
			}
			break;
		case 'r':
			opt->dsp_rate = atoi(optarg);
			break;
		case 'c':
			opt->dsp_ch = atoi(optarg);
			break;
		case 'D':
			debug_level = atoi(optarg);
			break;
		case 'X':
			opt->dsp_fd=atoi(optarg);
			break;
		case 'Y':
			opt->dsp_fmt=atoi(optarg);
			break;
		case 'd':
			constdelay=atoi(optarg);
			break;
		case '?':
		default:
			e_printf(helpmes, argv[0]);
			return 1;
		}
	}
	if(constdelay<0){
	  opt->constdelay_ts=-1;
	}else{
	  opt->constdelay_ts=constdelay*opt->emon_clk/1000;
	}
	return 0;
}


void audioplay_db_printobuf(int fd){
	audio_buf_info aubuf_info;

	ioctl(fd, SNDCTL_DSP_GETOSPACE, &aubuf_info);
	if(aubuf_info.bytes ==0 ){
		debug("buffer empty\n");

	}
	d1_printf("obuf fragments %d fragstotal %d" 
		 "fragsize %d bytes %d\n",
		 aubuf_info.fragments,
		 aubuf_info.fragstotal,
		 aubuf_info.fragsize,
		 aubuf_info.bytes);
}
