/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 *            
 * started:   2002/01/19
 *
 * $Id $
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include "def_value.h"
#include "pipe_hd.h"
#include "debug.h"
#include "timeval.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/*
  sprintf(tmp,argv_shcfmt,framefile1,framefile2);
  sh -c tmp
  for each frame
*/

/* 
   some limitation in this program
   1. only one fps for streams is acceptable (multi fps stream mix =>X)
   2. only one max frame size is acceptable  (waist memory => o)
   3. only one queue length for each stream
       is acceptable (this is because only one fps streams)
   4. Don't varify shc_fmt string is safty.


   process detail 
   1. argv parse and open all stream.
   2. until read the first frames for each stream ,
      repeat to call stream_update. 
   2.:1 
      if sync stream mode ,until the frames for each stream
      whoes timestamps are the same, are read,update and
      delete old frame. 
   3. set ts_base for all stream;
   4. set sigaciton for SIGCHLD to know the exit of child process.

   (start main loop)
   4. create pipe for all stream and shc;
   5. sprintf argv for "sh -c " and create execv arg2;
   6. fork;
    (child process)
     1.dup2 fd1(stdout => shc_out pipe)
     2. execv("/bin/sh",["sh" "-c" "argv for sh -c"])
   7. select 10ms
   8. check SIGCHLD. if exit =>10
   9. read frame to buf and write frame to pipe =>7

   10. close all pipe
   11. write emon message with payload length to pipe. 
       and stream_update =>4
       (if all next frame for each stream are read 
        or detected as lost frame.      ==> end update)
       (if recv=>display timeout occure ==> disable block update)
       (if updated frame => not update)
       (if sync mode && all stream have more than X frames in each queue,
       delete Y frames for each stream.
       )
   [stream_update]
    1.select(no wait) for passed fd(s).
    2.(for each stream)
      if queue have empth space,fill it.
      else 
        if    sync stream mode => delete new frame
	else  delete frames according to jitter,delay
	      and update ts_base.
 */

/* 
   queue struct
   (wait0->wait1->wait2->wait3->....waitX <--last_p
   or 
   wait1->wait2->wait3->....waitX <--last_p
   wait0)
   or
   (wait0->wait1->wait2->wait3->....waitX <--last_p
   or 
   wait0->wait1->wait2->wait3->....waitX->growing waitX+1 <--grow
   )
   interface
   1. update(q,fd)
   2. lock(q,&ret_mem &ret_ts)
   3. unlock(q)
   4. getnum(q)
   5. del(q,num)
   6. delonjitter(q,num)


 */
#define MAX(a,b) ((a)>(b)? (a):(b))
#define MIN(a,b) ((a)>(b)? (b):(a))

typedef struct _opt_t  opt_t;
typedef struct _noname_t  noname_t;

struct _opt_t{	/* set by getenv and/or getopt */
	int    emon_clk;
	int    emon_freq;
	int    ibuf_size;
	int    ibuf_num;
	int    obuf_size;
	
	char **exec;
	int    exec_len;
	char  *exec_sscanffd;
	char  *exec_fd;
};
struct _noname_t{	/* set by getenv and/or getopt */
	int      emon_clk;
	int      emon_freq;
	char    *ibuf;
	int      ibuf_size;
	char    *obuf;
	int      obuf_size;
	char **exec;
	int    exec_len;
	char  *exec_sscanffd;
	char  *exec_fd;
};

//static stat_t STAT;

int noname_getopt(int argc, char *argv[], opt_t *opt);
int noname_chkopt(opt_t *opt);
int noname_dumpopt(opt_t *opt);

int noname_init(opt_t *opt,noname_t **ret_noname);
int noname_main(noname_t *noname);


int
main(int argc, char *argv[])
{
	opt_t OPT;
	noname_t *noname;
	int   ret;

	ret=noname_getopt(argc, argv, &OPT);
	noname_dumpopt(&OPT);
	if (ret < 0){
		return 1;
	}
	debug_printf1("pass getopt\n");
#if 0
	ret=noname_chkopt(&OPT);
	noname_dumpopt(&OPT);
	if(ret>0){
		return 2;
	}
	debug_printf1("pass chkopt\n");
#endif
	ret=noname_init(&OPT,&noname);
	if(ret>0){
		return 3;
	}
	noname_main(noname);
	
	return 0;
}

int noname_init(opt_t *opt,noname_t **ret_noname){
	noname_t *noname;
	noname=malloc(sizeof(noname_t));
	noname->exec=opt->exec;

	*ret_noname=noname;
	return 0;
}


int noname_main(noname_t *noname){
	char *ibuf_top;
	char *ibuf;
	int   ibuf_size;
	int   ibuf_rest;
	int   ibuf_filled;
	char *obuf_top;
	char *obuf;
	int   obuf_size;
	int   obuf_rest;
	int   obuf_filled;
	ssize_t io_ret; 
	char  p_hd[PIPE_HEADER_LEN];
	pid_t child_pid;
	int   pipe_p2c[2]; /* p:write( pipe_p2c[1]) c:read (pipe_p2c[0]) */
	int   pipe_c2p[2]; /* p:read ( pipe_c2p[0]) c:write(pipe_c2p[1]) */
	int   nfds;
	int   select_ret;
	fd_set           fds_r,fds_w;
	struct timeval  selecttv;


	
	/* init */
	ibuf_size=1<<20;	
	obuf_size=1<<20;
	ibuf_top=malloc(ibuf_size);
	obuf_top=malloc(obuf_size);

	/* prebuf */
	debug_printf1("start prebuf\n");
	io_ret=pipe_blocked_read_block(STDIN_FILENO,p_hd,ibuf_top,ibuf_size);
	if(io_ret<0){
		e_printf("failt to read first frame.\n");
		return 1;
	}
	ibuf_filled=io_ret;

        /* start loop */
	debug_printf1("start loop\n");
loop_start:
	debug_printf2("write %d to child\n",ibuf_filled);
	/* exec */
	pipe(pipe_p2c);
	pipe(pipe_c2p);
	child_pid=fork();
	if(child_pid<0){
		e_printf("fail to fork for executing %s %s...\n"
			 ,noname->exec[0],noname->exec[1]);
		return 2;
	}
	if(child_pid==0){
		/* Child process */
		dup2(pipe_p2c[0],STDIN_FILENO);
		dup2(pipe_c2p[1],STDOUT_FILENO);
		close(pipe_p2c[1]);
		close(pipe_c2p[0]);
		execvp(noname->exec[0],noname->exec);
		/* if reach this line => execvp fail */
		e_printf("fail to execvp\n");
		_exit(0);
	}
	/* parent process */
	close(pipe_p2c[0]);
	close(pipe_c2p[1]);
	nfds=MAX(pipe_p2c[1],pipe_c2p[0]) +1;
	FD_ZERO(&fds_w);
	FD_ZERO(&fds_r);
	ibuf_rest=ibuf_filled;
	obuf_rest=obuf_size;
	ibuf=ibuf_top;
	obuf=obuf_top;
	for(;;){
		FD_CLR(pipe_p2c[1],&fds_w);	
		FD_CLR(pipe_c2p[0],&fds_r);
		if(ibuf_rest>0){
			FD_SET(pipe_p2c[1],&fds_w);
		}
		FD_SET(pipe_c2p[0],&fds_r);
		selecttv.tv_sec=0;
		selecttv.tv_usec=0;
		debug_printf3("start select\n");
		select_ret=select(nfds,&fds_r,&fds_w,NULL,NULL);
		debug_printf3("end select %d\n",select_ret);
		if(select_ret==-1){
			e_printf("select interrupted\n");
			continue;
		}
		if(FD_ISSET(pipe_c2p[0],&fds_r)){
			debug_printf3("read from child\n");
			io_ret=read(pipe_c2p[0],obuf,MIN(4096,obuf_rest));
			if(io_ret==0){
				/* child stoped */
				break;
			}
			if(io_ret<0){
				return 3;
			}
			obuf+=io_ret;
			obuf_rest-=io_ret;
		}
		if(FD_ISSET(pipe_p2c[1],&fds_w)){
			debug_printf3("write to child\n");
			io_ret=write(pipe_p2c[1],ibuf,MIN(4096,ibuf_rest));
			if(io_ret==0){
				/* child stoped */
				break;
			}
			if(io_ret<0){
				return 3;
			}
			ibuf+=io_ret;
			ibuf_rest-=io_ret;
			if(ibuf_rest==0){
				close(pipe_p2c[1]);
			}
		}
		debug_printf3("write rest%d read filled %d\n"
			      ,ibuf_rest,obuf_size-obuf_rest);		
	}
	/* child stoped */
	kill(child_pid,SIGTERM);
	if(ibuf_rest!=0){
		close(pipe_p2c[1]);		
	}
	close(pipe_c2p[0]);
	obuf_filled=obuf_size-obuf_rest;
	debug_printf2("write %d to stdout\n",obuf_filled);		
	pipe_set_length(p_hd,obuf_filled);
	pipe_blocked_write_block(STDOUT_FILENO,p_hd,obuf_top);

	/* fill ibuf */
	io_ret=pipe_blocked_read_block(STDIN_FILENO,p_hd,ibuf_top,ibuf_size);
	if(io_ret<0){
		e_printf("failt to read first frame.\n");
		return 1;
	}
	ibuf_filled=io_ret;

	{
		/* 
		   !Very important!
		   in high load environment,
		   if you do not wait for child exiting,
		   new child is created before old child exit.
		   In result ,
		   new child may be created before 
		   very old child exit, 
		   and exceed maxporc limit.
		*/
		int tmp;
		while(wait(&tmp)!=child_pid);
	}
	goto loop_start;
}


int noname_getopt(int argc, char *argv[], opt_t *opt){
	int ch;
	char *opts = "C:R:L:B:f:D:";
	char *helpmes =
		"usage : %s [options] delimiter exec ...\n"
		" options\n"
		" <general>\n"
		"  -C <n>   : clock base\n"
		"  -R <n>   : clock freq\n"
		"  -L <n>   : input msg size limit\n"
		"  -B <n>   : input msg buf limit\n"
		"  -f       : format of fd in exec ...[/dev/fd/msg%d]\n"
		" <debug>\n"
		"  -D <n>   : debug level\n"
		"\n";
	char *temp;

	opt->emon_clk = DEF_EMON_CLOCK;
	opt->emon_freq= DEF_EMON_FREQ;
	opt->exec_fd=NULL;
	opt->exec_sscanffd="/dev/fd/msg%d";
			/*  /dev/fd/9999*/
	opt->exec=NULL;
	opt->exec_len=0;

	temp=NULL;
	while ((ch = getopt(argc, argv, opts)) != -1){
		switch (ch){
		case 'f':
			opt->exec_sscanffd = optarg;
			break;
		case 'D':
			debug_level = atoi(optarg);
			break;
		default:
			e_printf(helpmes, argv[0]);
			return -1;
		}
	}
	if (argc - optind < 2){
		e_printf(helpmes, argv[0]);
		return -1;
	}
	opt->exec=argv+optind;
	opt->exec_len=argc-optind;
	return 0;
}


int noname_chkopt(opt_t *opt){
	int   idx;
	int   ret;
	int   chkopt_ret;
	int   pipe_fd;

	idx=0;
	chkopt_ret=1;
	while(idx<opt->exec_len){
		ret=sscanf(opt->exec[idx],opt->exec_sscanffd,&pipe_fd);
		debug_printf2("pipe_fd %d\n",pipe_fd);
		if(ret==1 && pipe_fd==1){
			opt->exec_fd=opt->exec[idx];
			chkopt_ret=0;
			break;
		}
		idx++;
	}
	return chkopt_ret;
}


int noname_dumpopt(opt_t *opt){
	int idx;

	debug_printf2("optdump\n");
	debug_printf2("exec=");
	for(idx=0;idx<opt->exec_len;idx++){
		d2_printf("\"%s\" ",opt->exec[idx]);
	}
	d2_printf("\n");
	debug_printf2("exec_sscanffd \"%s\"\n",opt->exec_sscanffd);
	debug_printf2("exec_fd       \"%s\"\n",opt->exec_fd);
	d2_printf("\n");
	return 0;
}
