/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KOMURA Takaaki <komura@astem.or.jp>
 *            KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp> 
 * started:   2001/06/26
 *
 * $Id: jpegplay.c,v 1.12 2001/11/16 08:02:54 kasamatu Exp $
 */

/*
 * test program for libjpeg + SDL
 */

#define REALTIME_PLAY
#define USE_YUV_OVERLAY		/* libjpeg outputs YCbCr, SDL handle YUV
				 * overlay */

#ifndef USE_YUV_OVERLAY		/* USE_RGB_SURFACE */
#define SDL_MASK_RGB_ORDER 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <string.h>
#include <signal.h>

#include <jpeglib.h>
#include "jpegbuf/bufio.h"

#include <SDL.h>
#include <SDL_timer.h>

#include "pipe_hd.h"
#include "timeval.h"
#include "pipeq.h"
#include "debug.h"
#include "def_value.h"

#define JPEG_BUF_MAX	(1024*100)	/* all jpeg date load in memory */

#ifndef RGB_PIXELSIZE		/* it's should be defined in jmorecfg.h */
#define RGB_PIXELSIZE 3
#endif
/* JPEG_YCbCr_PITCH means [YUV YUV YUV ..] for each pixels */
#define JPEG_YCbCr_PITCH RGB_PIXELSIZE


extern          GLOBAL(void) jpeg_buf_src(j_decompress_ptr, buf_t *);


static volatile int jpeg_has_error = 0;	/* loop flag in decoding jpeg */

enum disp_format {
	DF_YUV,
	DF_RGB,
	DF_BGR,
};

typedef struct opt_values {
	int             frame_rate;
	int		clock;
	int		freq;
	int             dsize;
	int             delay_limit; /*msec*/
	int             constdelay_ts;  /*msec*/
	double          mspf;	/* temprary millisec per frame for adjust */
	double          mspf_org;	/* original millisec per frame */
	int             f_disp;	/* flag */
	int             exec_opt_p;	/* exec options position in argv[] */
	int             output_width;
	int             output_height;
	int             sdl_mask_R, sdl_mask_G, sdl_mask_B, sdl_mask_A;
	enum disp_format dformat;
	char            title[255];
	int             loss_visual;
}               opt_t;



typedef struct statistics_values {
	struct timeval  start;
	int             frame_count;
	int             skip_count;
	int             wait_count;
	int             decode_usec;
	int             f_sigint;
}               stat_t;

static opt_t    OPT;
static stat_t   STAT;


/* prototype  */
static int      opt_etc(int argc, char *argv[], opt_t *opt);
void            statistics_print(stat_t * st);
void            sigint_quit(void);
void            sigint_trap(int sig);

pipeq_t        *pipeq; 

static
void 
wait4rtdisplay(struct timeval *tv_start,u_int64_t ts_limit){
	int64_t ts_diff;
	struct timeval  now;

	while(1){
	  gettimeofday(&now, NULL);		
	  ts_diff=timeval_sub_ts(&now,tv_start,OPT.clock);
	  if(ts_diff >= ts_limit){
	    break;
	  }
	  d3_printf("sleep.\n");
	  fd_can_read(STDIN_FILENO,1000);
	  pipeq_update(pipeq,STDIN_FILENO);	/* check new data */
	}
}

static
void 
wait4constdelay(u_int32_t ts_limit){
	struct timeval  now;
	u_int32_t       nowts;

	int  cnt=0;
	while(1){
	  gettimeofday(&now, NULL);		
	  nowts=timeval_2ts(&now,OPT.clock);
	  if((int32_t)(nowts-ts_limit)>=0){
	    break;
	  }
	  fd_can_read(STDIN_FILENO,1000);
	  pipeq_update(pipeq,STDIN_FILENO);	/* check new data */
	  cnt++;
	}
	if(cnt>0){
	  d2_printf("CONST DELAY ts%u <= r%u buf%d loop%d\n"
		   ,ts_limit&0xffffff,nowts&0xffffff
		   ,pipeq_filled(pipeq),cnt
		   );
	}
	  
}



void
my_jpeg_abort_decompress(j_common_ptr cinfo)
{
	d_printf("\nmy_jpeg_abort_decompress called, jpeg_has_error=%d\n",
		 jpeg_has_error);
	jpeg_has_error = 1;
	jpeg_abort_decompress((j_decompress_ptr) cinfo);
	exit(-1);
}



void 
ComplainAndExit(void)
{
	fprintf(stderr, "\nProblem: %s\n", SDL_GetError());
	exit(1);
}


static void jpeg_fillimg1color(SDL_Surface *img,u_int8_t *col,int col_len){
	int xl,xh,yl,yh,ystep,x,y;
	u_int8_t *p;

	if(col[0]<col[1]){
		yl=img->h/6 *2;
		yh=img->h/6 *3;
	}else if(col[1]<col[2]){
		yl=img->h/6 *3;
		yh=img->h/6 *4;
	}else{
		yl=img->h/6 *4;
		yh=img->h/6 *5;
	}
	ystep=1;
	if(col[0]<col[2]){
		xl=0;
		xh=img->w/2;
	}else{
		xl=img->w/2;
		xh=img->w-1;
	}

	y=yl;
	while(y<yh){
		p=img->pixels+img->pitch*y;
		for(x=xl;x<xh;x++){
			memcpy(p,col,col_len);
			p+=col_len;
		}
		y+=ystep;
	}
	return;
}
static int jpeg_blitimg2screen(SDL_Surface *screen,	
			       SDL_Surface *img){

	int ret;
	SDL_Rect        dstrect;

	if (SDL_MUSTLOCK(screen)) {
		if (SDL_LockSurface(screen) < 0) {
			ComplainAndExit();
		}
	}
	dstrect.x = 0;
	dstrect.y = 0;
	dstrect.w = img->w;
	dstrect.h = img->h;
	ret=0;
	if (SDL_BlitSurface(img, NULL, screen, &dstrect) < 0) {
		ret=1;
		goto fail;
	}
	SDL_UpdateRects(screen, 1, &dstrect);
	/* screen surface unlock */
 fail:		
	if (SDL_MUSTLOCK(screen)) {
		SDL_UnlockSurface(screen);
	}
	return ret;
}

struct jpeg_decompress_struct cinfo;
struct jpeg_error_mgr jerr;
int jpeg_dec_init(SDL_Surface **ret_screen,SDL_Surface **ret_img){
	buf_t           jpeg_src_buf;
	int             w, h, d;/* display size and depth(byte) */
	boolean         draw_screen;	/* draw screen surface or another
					 * surface */
	/* for jpeg library */

	/* for SDL */
	SDL_Surface    *screen;
	SDL_Surface    *sdl_image = NULL;

	/*  pipeq */
	void      *msg;
	u_int32_t  msgsize;
	u_int8_t   marker;
	u_int32_t  ts;			     
	void      *payload;
	u_int32_t  payloadsize;

	/* read jpeg image size */
 	cinfo.err = jpeg_std_error(&jerr); 
 	cinfo.err->error_exit = my_jpeg_abort_decompress; 
	jpeg_create_decompress(&cinfo);

	pipeq_lock(pipeq,&msg,&msgsize,&marker,&ts
			       ,&payload,&payloadsize);
	jpeg_src_buf.writecnt = payloadsize;
	jpeg_src_buf.readcnt = 0;
	jpeg_src_buf.buf = payload;
	jpeg_buf_src(&cinfo, &jpeg_src_buf);	

	jpeg_read_header(&cinfo, TRUE);
	w = cinfo.image_width;
	h = cinfo.image_height;
	d = cinfo.num_components;
	jpeg_abort_decompress(&cinfo);
	pipeq_unlock(pipeq);
	d_printf("\nJPEG info: image=(%d,%d), output=(%d,%d), Bpp=%d\n",
		 cinfo.image_width, cinfo.image_height, w, h, d);


	/* SDL init */
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		ComplainAndExit();
	}
	atexit(SDL_Quit);
	/* SDL setup / cleanup screen-surface */
	screen = SDL_SetVideoMode(w, h, 0, SDL_HWSURFACE);
	if (screen == NULL) {
		ComplainAndExit();
	}
	if (screen->format->BytesPerPixel == RGB_PIXELSIZE &&
	    screen->format->Rmask == OPT.sdl_mask_R &&
	    screen->format->Gmask == OPT.sdl_mask_G &&
	    screen->format->Bmask == OPT.sdl_mask_B) {
		draw_screen = TRUE;
	} else {
		draw_screen = FALSE;
	}
	d1_printf("\nSDL screen  info: bpp=%d, Bpp=%d, "
		  "R/G/B-mask=%06x/%06x/%06x, Direct=%s",
		  screen->format->BitsPerPixel, screen->format->BytesPerPixel,
		  screen->format->Rmask, screen->format->Gmask,
		  screen->format->Bmask,
		  draw_screen ? "ON" : "OFF");
	if (SDL_MUSTLOCK(screen) || draw_screen==FALSE) {
		sdl_image = SDL_CreateRGBSurface(
			SDL_HWSURFACE, w, h, 24,/* depth (bit per pixel) */
			OPT.sdl_mask_R, OPT.sdl_mask_G,
			OPT.sdl_mask_B, OPT.sdl_mask_A);
		*ret_img=sdl_image;
		d1_printf("\nSDL surface info: bpp=%d, Bpp=%d, "
			  "R/G/B-mask=%06x/%06x/%06x\n",
			  sdl_image->format->BitsPerPixel,
			  sdl_image->format->BytesPerPixel,
			  sdl_image->format->Rmask,
			  sdl_image->format->Gmask,
			  sdl_image->format->Bmask
			);

	} else {
		/* RGB_PIXELSIZE is defined in jmorecfg.h */
		*ret_img=screen;
	}
	*ret_screen=screen;
	return 0;
}

int jpeg_dec(u_int8_t *frame,u_int32_t size,SDL_Surface *img){
	buf_t           jpeg_src_buf;

	/* for debug */
	struct timeval  tv_tmp1, tv_tmp2;

	int             w, h, d;/* display size and depth(byte) */

	jpeg_src_buf.writecnt = size;
	jpeg_src_buf.readcnt = 0;
	jpeg_src_buf.buf = frame;
	jpeg_buf_src(&cinfo, &jpeg_src_buf);
	jpeg_read_header(&cinfo, TRUE);

	cinfo.output_width = img->w;
	cinfo.output_height = img->h;
	cinfo.out_color_space = JCS_RGB;	/* default */
//	cinfo.output_components = d;
	/* more fast decompression */
	cinfo.dct_method = JDCT_FASTEST;
	//cinfo.dct_method = JDCT_FLOAT;
	cinfo.do_fancy_upsampling = FALSE;

	jpeg_start_decompress(&cinfo);


	/* JPEG decode start */
	gettimeofday(&tv_tmp1, NULL);
	while (cinfo.output_scanline < cinfo.output_height &&
	       !jpeg_has_error) {
		JSAMPLE        *rgb_scanline;
		u_int8_t       *pixmap;
		pixmap=img->pixels;
		rgb_scanline = pixmap+cinfo.output_scanline*img->pitch;
		jpeg_read_scanlines(&cinfo, &rgb_scanline, 1);
	}
	jpeg_finish_decompress(&cinfo);
	gettimeofday(&tv_tmp2, NULL);
	STAT.decode_usec += timeval_diff_usec(&tv_tmp2, &tv_tmp1);
	/* JPEG decode ,frame => img finish */
	return 0;
}

int 
jpeg_display_rgb(int argc, char *argv[])
{
	int             screen_status; /* 0=general 1=blue 2=red */
	/* for jpeg library */
 	struct jpeg_decompress_struct cinfo; 
	struct jpeg_error_mgr jerr; 

	/* for SDL */
	SDL_Surface    *screen;
	SDL_Surface    *sdl_img = NULL;

	/* for visualize packet loss */
	void      *msg;
	u_int32_t  msgsize;
	u_int8_t   marker;
	void      *payload;
	u_int32_t  payloadsize;
	u_int8_t        pixel_blue[]={0,0,0xff};
	u_int8_t        pixel_red[]={0xff,0,0};
	u_int8_t        pixel_green[]={0,0xff,0};

	u_int32_t       tick_start, tick_now;

	/* for realtime play */
	struct timeval  tv_start,tv_now;
	int64_t         ts_display,ts_diff;
	u_int32_t       ts_nowblk,ts_lastblk=0;
	
	/* for debug */
	struct timeval  tv_tmp1, tv_tmp2;

	d1_printf("init start\n");
	jpeg_dec_init(&screen,&sdl_img);	
	d1_printf("init end\n");
//	sleep(2);

	STAT.skip_count = 0;
	STAT.wait_count = 0;
	tick_start = SDL_GetTicks();	/* get start time */
	gettimeofday(&STAT.start, NULL);
	STAT.f_sigint = 0;
	signal(SIGINT, sigint_trap);

	d1_printf("loop start\n");
	gettimeofday(&tv_start, NULL);
	ts_display=0;
	screen_status=0;
	for (STAT.frame_count = 1;; STAT.frame_count++) {
		if (STAT.f_sigint)
			sigint_quit();
		jpeg_has_error = 0;	/* reset flag */

		pipeq_update(pipeq,STDIN_FILENO);

		if(pipeq_filled(pipeq)== 0){ /* buffer empty */
			d1_printf("INFO:buffer empty\n");
			STAT.frame_count--;	/* no skip, no display */
			if(OPT.loss_visual && screen_status!=1){
				jpeg_fillimg1color(sdl_img,pixel_blue,3);
				jpeg_blitimg2screen(screen,sdl_img);
				screen_status=1;
			}
			d1_printf("INFO:fill%d empty%d\n"
				  ,pipeq_filled(pipeq)
				  ,pipeq_empty(pipeq));
			while(pipeq_filled(pipeq)<=pipeq_empty(pipeq)){
				pipeq_update(pipeq,STDIN_FILENO);
			}
			ts_display=0;
			gettimeofday(&tv_start, NULL);
			continue;
		} 
		d2_printf("lock frame\n");
		pipeq_lock(pipeq,&msg,&msgsize,&marker,&ts_nowblk
			   ,&payload,&payloadsize);
		d2_printf("unlock frame,start dec\n");
		jpeg_dec(payload,payloadsize,sdl_img);
		d2_printf("end dec,unlock frame\n");
		pipeq_unlock(pipeq);
		d2_printf("end unlock,del1\n");
		pipeq_rm(pipeq,1);
		d2_printf("del1 end,update\n");
		pipeq_update(pipeq,STDIN_FILENO);
#ifdef REALTIME_PLAY
		if(OPT.constdelay_ts<0){
		  wait4rtdisplay(&tv_start,ts_display);

		  if(ts_display==0){
		    ts_diff=OPT.freq;
		  }else{
		    ts_diff=(u_int32_t)(ts_nowblk-ts_lastblk);
		  }
		  ts_lastblk=ts_nowblk;

		  if(ts_diff<OPT.freq){
		    e_printf("INFO:reset TS interval\n");
		  }
		  if(ts_diff>OPT.freq){
		    d2_printf("blue back start for %u TS\n",(u_int32_t)ts_diff-OPT.freq);
		    ts_display+=(ts_diff-OPT.freq);
		    if(OPT.loss_visual && screen_status!=1){
			    jpeg_fillimg1color(sdl_img,pixel_blue,3);
			    jpeg_blitimg2screen(screen,sdl_img);
			    screen_status=2;
		    }
		    wait4rtdisplay(&tv_start,ts_display);
		    d2_printf("blue back end.\n");
		  }else{
		    d3_printf("no blue back .\n");
		  }
		  ts_display+= OPT.freq;
		}else{
		  wait4constdelay(ts_nowblk+(u_int32_t)OPT.constdelay_ts);
		}
#endif
		pipeq_update(pipeq,STDIN_FILENO);		
                if(pipeq_empty(pipeq)==0){
                        /* need more cpu power */
                        if(OPT.loss_visual){
				jpeg_fillimg1color(sdl_img,pixel_red,3);
                        }
                        d1_printf("INFO:buffer full ,skip fr%d/2\n",
                                  pipeq_filled(pipeq));
                        STAT.skip_count+=pipeq_filled(pipeq)/2;
                        STAT.frame_count+=pipeq_filled(pipeq)/2;
                        pipeq_rm(pipeq,pipeq_filled(pipeq)/2);
                        ts_display=0;
                        gettimeofday(&tv_start, NULL);
                }
		d2_printf("play frame .\n");
		jpeg_blitimg2screen(screen,sdl_img);
		screen_status=0;
	}			/* loop for all jpeg file */
	d1_printf("end play\n");
	STAT.frame_count--;	/* because count from 1 */
	tick_now = SDL_GetTicks();

	if (sdl_img!=screen) {
		SDL_FreeSurface(sdl_img);
	}
	jpeg_destroy_decompress(&cinfo);
	statistics_print(&STAT);
	return 0;
}

#define STDIN 0

int 
main(int argc, char *argv[])
{
	int buf_num;

	if (opt_etc(argc, argv, &OPT) == -1) {
		return -1;	/* quit */
	}
	buf_num=(OPT.frame_rate*OPT.delay_limit)/1000;
	if(buf_num<=2)
		buf_num=2;
	pipeq_create(OPT.dsize,buf_num,0,&pipeq);
	d1_printf("bufsize %d bufnum%d\n",OPT.dsize,buf_num);
	while(pipeq_filled(pipeq)<=buf_num/2){
		pipeq_update_blk(pipeq,STDIN);
	}
	pipeq_rm(pipeq,1);

	switch (OPT.dformat) {
	case DF_YUV:
		d1_printf("YUV format");
		fprintf(stderr, "cannot support YUV output format\n");
		break;
	case DF_RGB:
		d1_printf("RGB format");
		OPT.sdl_mask_R = 0xff0000;
		OPT.sdl_mask_G = 0x00ff00;
		OPT.sdl_mask_B = 0x0000ff;
		OPT.sdl_mask_A = 0;
		jpeg_display_rgb(argc, argv);
		break;
	case DF_BGR:
		d1_printf("BGR format");
		OPT.sdl_mask_R = 0x0000ff;
		OPT.sdl_mask_G = 0x00ff00;
		OPT.sdl_mask_B = 0xff0000;
		OPT.sdl_mask_A = 0;
		jpeg_display_rgb(argc, argv);
		break;
	default:
		break;
	}
	fprintf(stdout, "\n");
	return 0;
}



static int
opt_etc(int argc, char *argv[], opt_t *opt)
{
	int             ch;
	int             loop_flag = 1;
	char           *temp;
	int             constdelay;
	char           *opts = "r:nD:f:g:T:bs:W:d:";
	char           *helpmes =
	"usage : %s [-r <framerate>] [-D <debug level>]"
	" [-f <YUV|RGB|GRB>] [-n]  [-T <window title>]"
	" [-g <xsize>x<ysize> (not supported)] [-b]"
	"\n"
	" -n : no display (decode only)\n"
	" -b : if frame loss,fill screen with blue [require -d BRG]\n"
	" -s <n>   : max frame size\n"
	" -W <n>   : max delay # msec\n"
	" -d <n>   : constant delay # msec\n"
		;
	int             w, h;

	opt->clock = DEF_EMON_CLOCK;
	opt->freq = DEF_EMON_FREQ;
	opt->f_disp = 1;
	opt->exec_opt_p = 0;
	opt->output_width = 0;
	opt->output_height = 0;
	opt->loss_visual = 0;
#if 0
	opt->dformat = DF_YUV;
#else
	opt->dformat = DF_BGR;
#endif
	opt->dsize   =DEF_JPEGPLAY_DSIZE;
	opt->delay_limit=DEF_JPEGPLAY_DELAY_LIMIT;
	constdelay=-1;
	strncpy(opt->title, "jpegplay", sizeof(opt->title) - 1);

	if ((temp = getenv("EMON_CLOCK")) != NULL)
		opt->clock = atoi(temp);
	if ((temp = getenv("EMON_FREQ")) != NULL)
		opt->freq = atoi(temp);
	if ((temp = getenv("JPEGPLAY_DSIZE")) != NULL)
		opt->dsize = atoi(temp);
	if ((temp = getenv("JPEGPLAY_DELAY_LIMIT")) != NULL)
		opt->delay_limit = atoi(temp);

	while ((ch = getopt(argc, argv, opts)) != -1 && loop_flag == 1) {
		switch (ch) {
		case 'r':
			opt->frame_rate = atoi(optarg);
			break;
		case 'D':
			debug_level = atoi(optarg);
			break;
		case 'f':	/* display format */
			if (strcasecmp(optarg, "YUV") == 0) {
				opt->dformat = DF_YUV;
			} else if (strcasecmp(optarg, "RGB") == 0) {
				opt->dformat = DF_RGB;
			} else if (strcasecmp(optarg, "BGR") == 0) {
				opt->dformat = DF_BGR;
			} else {
				opt->dformat = DF_YUV;
				fprintf(stderr, "invalid option in -d option");
			}
			break;
		case 'g':	/* geometry () */
			sscanf(optarg, "%dx%d", &w, &h);
			if (!((0 < w && w <= 1280) && (0 < h && h <= 1024))) {
				break;
			}
			opt->output_width = w;
			opt->output_height = h;
			break;
		case 'n':
			opt->f_disp = 0;
			break;
		case 'b':
			opt->loss_visual = 1;
			break;
		case 's':
			opt->dsize = atoi(optarg);
			break;
		case 'd':
			constdelay = atoi(optarg);
			break;
		case 'W':
			opt->delay_limit = atoi(optarg);
			break;
		case 'T':	/* Window Title */
			strncpy(opt->title, optarg, sizeof(opt->title) - 1);
			break;
		default:
			fprintf(stderr, helpmes, argv[0]);
			return -1;
		}
	}

	if (opt->freq < 0){
		e_printf("freq (-F or EMON_FREQ) value must be over 0\n");
		return -1;
	}
	opt->frame_rate = opt->clock / opt->freq;

	if (opt->frame_rate == 0) {
		opt->frame_rate=30;
		opt->mspf_org = 1000.0 / 30;
	} else {
		opt->mspf_org = 1000.0 / opt->frame_rate;
	}
	opt->mspf = opt->mspf_org;
	if(constdelay<0){
	  opt->constdelay_ts=-1;
	}else{
	  opt->constdelay_ts=constdelay*opt->clock/1000;
	  e_printf("constdelay_ts %u\n",opt->constdelay_ts);
	}

	return 0;
}






void
sigint_trap(int sig)
{
	STAT.f_sigint = 1;
	pipe_exit_req();
}


void
sigint_quit(void)
{
	int             pid = getpid();

	printf("\n------------------ SIGINT (playjpeg/pid=%5d)\n", pid);
	statistics_print(&STAT);

	SDL_Quit();
	exit(-1);
}

void
statistics_print(stat_t * st)
{
	struct timeval  now;
	int             diff_msec;
	int             disp_count = st->frame_count - st->skip_count;

	gettimeofday(&now, NULL);
	diff_msec = (now.tv_sec - st->start.tv_sec) * 1000;
	diff_msec += (now.tv_usec - st->start.tv_usec) / 1000;

	fprintf(stdout,
		"\ndisplay: %d(skip: %d, wait: %d), "
		"%.2f f/s, (include skip %.2f f/s) decode %d us/f",
		disp_count,
		st->skip_count,
		st->wait_count,
	    ((float) (st->frame_count - st->skip_count) * 1000) / diff_msec,
		((float)st->frame_count * 1000) / diff_msec,
		st->decode_usec / disp_count
		);
}
