/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2001/0?/??
 *
 * $Id$
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#include "debug.h"
#include "pipe_hd.h"
#include "string.h"


#define MAX(a,b) (a>b? a:b)
#define MIN(a,b) (a>b? b:a)
       


typedef struct _pipecapt_t{
	int                emon_clk; 
	int                emon_freq;
        int                plen;
}pipecapt_t;


u_int32_t pipecapt_getopt(int argc, char *argv[],pipecapt_t *opt);
u_int32_t pipecapt_main(pipecapt_t *opt);

int main(int argc, char **argv)
{
	pipecapt_t opt;
	u_int32_t  ret;

	ret=pipecapt_getopt(argc, argv, &opt);
//	pipecapt_dumpopt(&opt);
	if (ret > 0){
		return 1;
	}
	debug_printf1("pass getopt\n");
	pipecapt_main(&opt);
	return 0;

}


u_int32_t pipecapt_main(pipecapt_t *opt){
	char      *buf;
	char       p_hd[PIPE_HEADER_LEN];
	u_int32_t  ts;
	ssize_t    io_ret;

	
	buf=malloc(opt->plen);
	ts=0;
	bzero(p_hd,PIPE_HEADER_LEN);
	pipe_set_version(p_hd,1);
	pipe_set_marker(p_hd,1);
        pipe_set_length(p_hd,opt->plen);

	while(1){
		io_ret=fd_read_ignoreEOF(STDIN_FILENO,buf,opt->plen);
		pipe_set_timestamp(p_hd, ts);
		pipe_blocked_write_block(STDOUT_FILENO,p_hd,buf);
		ts+=opt->emon_freq;
	}
}


u_int32_t pipecapt_getopt(int argc, char *argv[],pipecapt_t *opt){
	char            c;
	char           *tmpstr;	/* for getenv */

	char           *opts = "C:R:L:h";
	char           *helpmes =
	"usage : %s [options ...]\n"
	" options\n"
	"  -C <num> : EMON clock[44100]\n"
	"  -R <num> : EMON frequency[1470]\n"
	"  -L <num> : emon payload length limit[160]\n"
		;
	opt->emon_clk  = 44100;
	opt->emon_freq = 1470;
	opt->plen      = 160;

	while ((c = getopt(argc, argv, opts)) != -1) {
		switch (c) {
		case 'C':
			opt->emon_clk = atoi(optarg);
			break;
		case 'R':
			opt->emon_freq = atoi(optarg);
			break;
		case 'L':
			opt->plen = atoi(optarg);
			break;
		case '?':
		default:
		        e_printf(helpmes, argv[0]);
			return 1;
		}
	}
	return 0;
}
