/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): FUJIKAWA Kenji <fujikawa@real-internet.org>
 *            KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2001/08/12
 *
 * $Id: pipeinfo.c,v 1.2 2001/10/03 11:34:54 kasamatu Exp $
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "pipe_hd.h"
#include "rtp_hd.h"

#define MAX(a,b) ((a)>(b)? (a):(b))
#define MIN(a,b) ((a)>(b)? (b):(a))

void usage();

int
main(int argc, char *argv[])
{
	char ch;
#define EPI_BUFSIZE 4096 /* EMON PIPEINFO BUFSIZE */
	char buf[EPI_BUFSIZE];
	pipe_hd_t *phdr = (pipe_hd_t*)buf;
	rtp_hd_t *rhdr = (rtp_hd_t*)(phdr+1);
	int len, len2;
	ssize_t io_ret;
	//	fd_set rfds;

	while ((ch = getopt(argc, argv, "h")) != -1) {
		switch (ch) {
		case 'h':
		default:
			usage();
		}
	}
	/*fcntl(0, F_SETFL, fcntl(0, F_GETFL)|O_NONBLOCK);*/
	for (;;) {
#if 0
		FD_ZERO(&rfds);
		FD_SET(0, &rfds);
		select(0+1, &rfds, NULL, NULL, NULL);
#endif
		len = fd_read_ignoreEOF(0, buf, PIPE_HEADER_LEN);
		if (len != PIPE_HEADER_LEN) {
			fprintf(stderr, "EMON header error (len=%d)\n", len);
			exit(1);
		}
		len = ntohl(phdr->vmlen)&0x00ffffff;
		len2=MIN(len,sizeof(rtp_hd_t));
		io_ret=fd_read_ignoreEOF(0,buf + PIPE_HEADER_LEN, len2);
		if (len2 != io_ret) {
			fprintf(stderr, 
				"size is illeagal len=%d, len2=%d\n", 
				io_ret, len2);
			exit(1);
		}
		fprintf(stderr, 
			"len=%05d, vmlen=0x%08x, ts=%010d\n"
			"vpxc=0x%02x, mp=0x%02x, seq=%05d, ts=%010d\n",
			len, ntohl(phdr->vmlen), ntohl(phdr->ts),
			rhdr->vpxc, rhdr->mp,ntohs(rhdr->seq),ntohl(rhdr->ts));
		io_ret=fd_lseek_ignoreEOF(0,len-len2);
		if(io_ret != (len-len2)){
			exit(1);
		}

	}

	return (0);
}

void usage()
{
	fprintf(stderr, "usage: pipeinfo\n");
	exit(1);
}
