/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2002/01/22
 *
 * $Id$
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "pipe_hd.h"
#include "debug.h"

#include "pipeq.h"


/* TODO
   u_int32_t ts_clk;
   int32_t   ts_base; // only for delay  
   int32_t   delay;   // only for jitter
   u_int32_t jitter; 

   jitter progress
   get new_ts in EMON message
   new_delay =(wclk_ts)-(new_ts+ts_base)	
   new_jitter= abs((delay16>>16)-new_delay);
   delay16 -= delay16 >>4;
   delay16 += (new_delay<<16)>>4;
   jitter16 -= jitter16 >>4;
   jitter16 += (new_jitter<<16)>>4;
   ts_base+=delay16>>16;
   delay16=delay16&(1<<16 -1);
*/


typedef struct _frame_t frame_t;

#define MAX(a,b) ((a)>(b)? (a):(b))
#define MIN(a,b) ((a)<(b)? (a):(b))

struct _pipeq_t{
	u_int8_t  ignore_marker; /* !0:one EMON message is one frame */

	u_int32_t size;
	struct incoming_t{
		frame_t  *frame;	
		u_int32_t rest;
		u_int32_t size;
		u_int8_t  *now;
		u_int32_t ts;
		u_int8_t  marker;  /* 0:no marker 1:marker */
		u_int8_t  append; /* 0:loss 1:append*/
		u_int8_t  tscheck;/* 0:start new frame 1:ts check */
	}incoming;
	u_int32_t num;
	u_int32_t fillednum;
	u_int32_t emptynum;

	frame_t *lock;	
	frame_t *first;		/* oldest filled */
	frame_t *last;		/* newest filled */
	frame_t *empty;	
};
struct _frame_t{
	frame_t   *next;	/* to newer(the newest frame,next=NULL */
	u_int32_t size;
	u_int8_t  buf[1];
};
#define FRAME_T_SIZE(msg_len) (sizeof(frame_t)-1+msg_len)



u_int32_t pipeq_create      (u_int32_t  framesize
			     ,u_int32_t framenum
			     ,u_int32_t ignore_marker
			     ,pipeq_t   **ret_pipeq)
/* 
   ignore_marker==0 ,marker is frame delimiter
   ignore_marker!=0 ,marker is not frame delimiter

   ret=pipeq_create(framesize,framenum,1,&pipeq);
   if(ret){/# need more memory #/
   }else{/# succeed #/
   }
 */
{
	pipeq_t   *pipeq;
	u_int32_t size;
	u_int32_t cnt;
	u_int8_t  *p;
	frame_t  *frame;
	size=0;
	size+=sizeof(pipeq_t);
	size+=FRAME_T_SIZE(framesize)*framenum;
	p=(u_int8_t*)malloc(size);
	if(p==NULL){
		return 1;
	}
	/* memory distribution */
	pipeq=(pipeq_t*)p;
	p+=sizeof(pipeq_t);
	pipeq->empty=(frame_t*)p;
	frame=(frame_t*)p;
	for(cnt=0;cnt<framenum;cnt++){
		frame=(frame_t*)p;
		p+=FRAME_T_SIZE(framesize);
		frame->next=(frame_t*)p;
	}
	frame->next=NULL;

        /* pipeq parameter initialize */
	pipeq->ignore_marker=ignore_marker;
	pipeq->size=framesize;
	pipeq->incoming.rest=0;
	pipeq->incoming.marker=1;
	pipeq->num=framenum;
	pipeq->fillednum=0;
	pipeq->emptynum =framenum;
	pipeq->lock=NULL;
	pipeq->first=NULL;
	pipeq->last=NULL;
	/*pipeq->empty*/

	*ret_pipeq=pipeq;
	return 0;
}


static void pipeq_update_dropincoming(pipeq_t *pipeq){
	struct incoming_t *in;     
	in=&(pipeq->incoming);

	/* incoming to empty */
	in->frame->next=pipeq->empty;
	pipeq->empty=in->frame;
	pipeq->emptynum++;
}
/* 
   ret=pipeq_update(pipeq,fd);
   switch(ret){
   case 0:/# ERROR:io#/
   case 1:/# ERROR:no empty frame#/
   case 2:/# ERROR:lost new frame because it is too large.#/
   case 3:/# update succeeded #/
   case 4:/# update succeeded and have read a frame #/
   case 5:/# have read a frame and empty frame has been nothing#/
   }
 */
static u_int32_t pipeq_update_newframe(pipeq_t *pipeq ,int fd,
				       int32_t *ret_loop){
	ssize_t            io_ret;
	char              *p_hd;
	char               p_hd_mem[PIPE_HEADER_LEN];
	struct incoming_t *in;     
	u_int32_t          ret;

        /* assert in->rest==0 */
	ret=3;
	in=&(pipeq->incoming);

	/* read new packet */
	if(in->marker){ /* no check ts,first packe of a frame */
		if(pipeq->empty==NULL){
			*ret_loop=0;
			return 1;
		}
		in->frame=pipeq->empty;
		pipeq->empty=pipeq->empty->next;
		pipeq->emptynum--;

		in->now=in->frame->buf;
		in->size=0;
		in->append=1;
		p_hd=in->now;
		in->now+=PIPE_HEADER_LEN;
	}else{/* check ts,not a first packe of a frame */
		p_hd=p_hd_mem;
	}
	io_ret=fd_read_ignoreEOF(fd,p_hd,PIPE_HEADER_LEN);
	if(io_ret!=PIPE_HEADER_LEN){
		goto io_err;
	}
	in->rest=pipe_get_length(p_hd);
	if(in->marker==0 && in->ts!=pipe_get_timestamp(p_hd)){
		in->append=0;
	}
	if(in->rest==0){/* lost packet */
		in->append=0;
	}		
	in->ts=pipe_get_timestamp(p_hd);
	if(pipeq->ignore_marker){
		in->marker=1; /* last MSG of frame */
	}else{
		in->marker=pipe_get_marker(p_hd);
	}
	if(in->append==1 && 
	   in->rest >pipeq->size - in->size - PIPE_HEADER_LEN){
		/* too large packet */
		in->append=0;
		*ret_loop=0;
		return 2;
	}
	if(in->marker==1 && in->rest==0){
		pipeq_update_dropincoming(pipeq);
	}
	*ret_loop=1;
	return 3;
io_err:
	/* incoming to empty */
	in->frame->next=pipeq->empty;
	pipeq->empty=in->frame;
	pipeq->emptynum++;
	*ret_loop=0;
	return 0;
}


static u_int32_t pipeq_update_growframe(pipeq_t *pipeq ,int fd,
					int32_t *ret_loop){
	ssize_t            io_ret;
	ssize_t            io_len;
	struct incoming_t *in;     
	u_int32_t          ret;

	ret=3;
	in=&(pipeq->incoming);
        /* assert in->rest>=1 */

	/* append msg to buf or drop msg */
	if(in->append){
		debug_printf3("pipeq_read\n");
		io_len=MIN(0x1000,in->rest);
		io_ret=read(fd,in->now,io_len);
		if(io_ret <0){
			debug_printf3("pipeq_read fail\n");
			goto io_err;
		}
		in->now+=io_ret;
		in->size+=io_ret;
		in->rest-=io_ret;
	}else{
		debug_printf3("pipeq_seek\n");
		io_len=MIN(0x1000,in->rest);
		io_ret=fd_lseek_ignoreEOF(fd,io_len);
		if(io_ret != io_len){
			debug_printf3("pipeq_seek fail\n");
			goto io_err;
		}
		in->rest-=io_len;
	}

	ret=3;
	if(in->rest==0 && in->marker){
		/* rename incoming to filled or empty */
		if(in->append){
			/* incoming to filled */
			d2_printf("pipeq_frame_read,filled%d,size%d\n"
				,pipeq->fillednum,in->size);
			pipe_set_length(in->frame->buf,in->size);
			in->frame->next=NULL;
			if(pipeq->last==NULL){
				pipeq->first=in->frame;
			}else{
				pipeq->last->next=in->frame;
			}
			pipeq->last=in->frame;

			pipeq->fillednum++;
			if(pipeq->empty==NULL){
				ret=5;
			}else{
				ret=4;
			}
		}else{
			/* incoming to empty */
			pipeq_update_dropincoming(pipeq);
		}
	}
	*ret_loop=1;
	return ret;
io_err:
	/* incoming to empty */
	in->frame->next=pipeq->empty;
	pipeq->empty=in->frame;
	pipeq->emptynum++;
	*ret_loop=0;
	return 0;
}

u_int32_t pipeq_update      (pipeq_t    *pipeq
			     ,int fd){
	u_int32_t   ret,loop;

	ret=3;
	while(fd_can_read(fd,0)){
		if(pipeq->incoming.rest==0){
			ret=pipeq_update_newframe(pipeq,fd,&loop);
		}else{
			ret=pipeq_update_growframe(pipeq,fd,&loop);
		}
		d3_printf("update ret=%u\n",ret);
                if(!loop)
			break;
	}
	d3_printf("updated ret=%u\n",ret);
	return ret;
}

u_int32_t pipeq_update_blk   (pipeq_t    *pipeq
			      ,int fd){
	u_int32_t blk;
	blk=1;
	while(1){
		blk=fd_can_read(fd,10000);
		if(blk)
			break;
	}
	return (pipeq_update(pipeq,fd));
}

/* 
   ret=pipeq_lock(pipeq,&frame,&frame_size)
   ret==0 =>locked
   ret==1 =>fail (,this call treated as "pipeq_unlock; this_call;")
 */
u_int32_t pipeq_lock        (pipeq_t     *pipeq /* !=NULL */
			     ,void      **ret_msg 
			     ,u_int32_t  *ret_size
			     ,u_int8_t   *ret_marker
			     ,u_int32_t  *ret_ts			     
			     ,void      **ret_payload
			     ,u_int32_t  *ret_payloadsize){
	frame_t *frame;
	if(pipeq->lock!=NULL){
		return 1;
	}
	if(pipeq->first==NULL){
		return 2;
	}
	frame=pipeq->first;
	pipeq->first=pipeq->first->next;
	if(pipeq->first==NULL){
		pipeq->last=NULL;
	}
	pipeq->lock=frame;
	pipeq->fillednum--;
	*ret_msg=frame->buf;
	*ret_size=pipe_get_length(frame->buf)+PIPE_HEADER_LEN;
	*ret_marker=pipe_get_marker(frame->buf);
	*ret_ts=pipe_get_timestamp(frame->buf);
	*ret_payload=frame->buf+PIPE_HEADER_LEN;
	*ret_payloadsize=pipe_get_length(frame->buf)+PIPE_HEADER_LEN;

	return 0;
}
/* 
   ret=pipeq_unlock(pipeq)
   ret==0 =>unlocked
   ret==1 =>frame is not locked 
 */
u_int32_t pipeq_unlock      (pipeq_t    *pipeq){
	if(pipeq->lock==NULL){
		return 1;
	}
	/* update filled list */
	pipeq->lock->next=pipeq->first;
	pipeq->first=pipeq->lock;
	if(pipeq->last==NULL){
		pipeq->last=pipeq->lock;
	}
	pipeq->lock=NULL;
	pipeq->fillednum++;
	d2_printf("buf%d fi%d em%d \n",
		  pipeq->num,pipeq->fillednum,pipeq->emptynum);
	return 0;
}

/* 
   locked frame is not deleted.
   ret=pipeq_rm(pipeq,num);
   ret == 0 =>succeeded 
   ret == 1 =>fail (this call treated as "pipeq_rm(q,pipeq_filled))")
 */
u_int32_t pipeq_rm          (pipeq_t    *pipeq
			     ,u_int32_t  num){

	u_int32_t qlen;
	u_int32_t cnt;
	u_int32_t ret;
	frame_t   *frame,*newempty;
	static int cntX=0;
	cntX++;

	qlen=pipeq_filled(pipeq);
	if (num>qlen){
		num=qlen;
		ret=1;
	}else{
		ret=0;
	}
	if(num==0){
		return ret;
	}

        /* update pipeq->{first,last} */
	d2_printf("update first,%d\n",cntX);
	newempty=pipeq->first;
	if(qlen==num){
		frame=pipeq->last;
		pipeq->first=NULL;
		pipeq->last=NULL;
	}else{
		frame=pipeq->first;
		for(cnt=1;cnt<num;cnt++){
			frame=frame->next;
		}
		pipeq->first=frame->next;
	}
	/* update empty */
	d2_printf("update empty\n");
	frame->next=pipeq->empty;
	pipeq->empty=newempty;
	pipeq->emptynum +=num;
	pipeq->fillednum-=num;
	return 0;
}

/* 
   qlen=pipeq_filled(pipeq);
   ret == number of filled frames !!excluding!! locked frame.
 */
u_int32_t pipeq_filled       (pipeq_t    *pipeq){
	return(pipeq->fillednum);
}
/* 
   qlen=pipeq_empty(pipeq);
   ret == number of empty frames including growthing frame. 
 */
u_int32_t pipeq_empty        (pipeq_t    *pipeq){
	return(pipeq->emptynum);
}

/* 
   ret=pipeq_destroy(pipeq);
   ret == 0 succeeded
   ret == 1 fail 
 */
u_int32_t pipeq_destroy      (pipeq_t    *pipeq){
	free(pipeq);
	return 0;
}

/* 
   qlen=pipeq_len(pipeq);
   ret == number of queue length (passed to pipeq_create).
 */
u_int32_t pipeq_framenum       (pipeq_t    *pipeq){
	return pipeq->num;
}
/* 
   qlen=pipeq_len(pipeq);
   ret == number of queue length (passed to pipeq_create).
 */
u_int32_t pipeq_framesize      (pipeq_t    *pipeq){
	return pipeq->size;
}
