/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi
 * started:   2002/01/22
 *
 * $Id$
 */

#ifndef _pipeq_h_
#define _pipeq_h_

#include <sys/types.h>
typedef struct _pipeq_t pipeq_t;


u_int32_t pipeq_create      (u_int32_t  framesize,
			     u_int32_t framenum,
			     u_int32_t ignore_marker,
			     pipeq_t   **ret_pipeq);
/* 
   ignore_marker==0 ,maker is frame delimiter
   ignore_marker!=0 ,maker is not frame delimiter

   ret=pipeq_create(framesize,framenum,1,&pipeq);
   if(ret){/# need more memory #/
   }else{/# succeed #/
   }
 */

u_int32_t pipeq_update_blk   (pipeq_t    *pipeq,
			      int fd);
u_int32_t pipeq_update      (pipeq_t    *pipeq,
			     int fd);
/* 
   ret=pipeq_update(pipeq,fd);
   switch(ret){
   case 0:/# ERROR:io#/
   case 1:/# ERROR:no empty frame#/
   case 2:/# ERROR:lost new frame because it is too large.#/
   case 3:/# update succeeded #/
   case 4:/# update succeeded and have read a frame #/
   case 5:/# have read a frame and threre is no buf for incoming msgs #/
   }
 */
#if 0
u_int32_t pipeq_lock        (pipeq_t     *pipeq ,/* !=NULL */
			     u_int32_t    idx, /* 0,1,...,pipq_filled()-1 */
			     void       **ret_msg, 
			     u_int32_t   *ret_size,
			     void       **ret_key);
#else

u_int32_t pipeq_lock        (pipeq_t     *pipeq /* !=NULL */
			     ,void      **ret_msg 
			     ,u_int32_t  *ret_size
			     ,u_int8_t   *ret_marker
			     ,u_int32_t  *ret_ts			     
			     ,void      **ret_payload
			     ,u_int32_t  *ret_payloadsize);
#endif
/*
   ret=pipeq_lock(pipeq,0,&frame,&frame_size,&frame_key)
   ret==0 =>locked
   ret==1 =>fail (,this call treated as "pipeq_unlock; this_call;")
 */
#if 0
u_int32_t pipeq_unlock      (pipeq_t    *pipeq,
			     void       *key);
#else
u_int32_t pipeq_unlock      (pipeq_t    *pipeq);
#endif
/* 
   ret=pipeq_unlock(pipeq)
   ret==0 =>unlocked
   ret==1 =>frame is not locked 
 */

u_int32_t pipeq_rm          (pipeq_t    *pipeq
			     ,u_int32_t  num);
/* 
   locked frame is not deleted.
   ret=pipeq_rm(pipeq,num);
   ret == 0 =>succeeded 
   ret == 1 =>fail (this call treated as "pipeq_rm(q,pipeq_filled))")
 */

u_int32_t pipeq_filled       (pipeq_t    *pipeq);
/* 
   ret=pipeq_filled(pipeq);
   ret==number of filled frames including locked frame.
 */
u_int32_t pipeq_empty        (pipeq_t    *pipeq);
/* 
   ret=pipeq_empty(pipeq);
   ret == number of empty frames including growthing frame. 
 */

u_int32_t pipeq_destroy      (pipeq_t    *pipeq);
/* 
    free memory allocated for pipeq_t and pipeq->frames.
   ret=pipeq_destroy(pipeq);
   ret == 0 succeeded
   ret == 1 fail 
 */

u_int32_t pipeq_framenum       (pipeq_t    *pipeq);
/* 
   qlen=pipeq_len(pipeq);
   ret == number of queue length (passed to pipeq_create).
 */
u_int32_t pipeq_framesize      (pipeq_t    *pipeq);
/* 
   qlen=pipeq_len(pipeq);
   ret == number of queue length (passed to pipeq_create).
 */
void            decoder_buf_rm(int num);

#endif
