/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp> 
 *            
 * started:   2002/01/24
 *
 * $Id$
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "string.h"


/* 
   the extension of the standard "strsep".
   this function add # of separated str and wheather
   # of found delim in strp[0] is more than strplen or not
   , to "strsep".

   ret=string_sep(strp,":-/<>",3,&strnum);
   switch(ret){
   case 0:/# succeed #/
   case 1:/# too many delim found #/
   } 
 */
int string_sep(char **strp,const char *delim,int strplen,int *ret_strplen){
	int   num;
	char *s;
	int   ret;

	num=1;
	s=strp[0];
	while(1){
		s=strpbrk(s,delim);
		if(s==NULL){
			ret=0;
			break;
		}
		*s='\0';
		s++;
		if(num==strplen){
			if(*s!='\0'){
				ret=1;
				break;
			}else{
				ret=0;
				break;
			}
		}
		strp[num]=s;
		num++;
	}
	*ret_strplen=num;
	return ret;
}


/* 
   MEMO:
   if recv packet by multicast. src_addr is multicast address
   if_addr is a addr this host have.
   str="src_addr/src_port/if_addr/ipvar"
 */
int string_sep4udpiprecv(char *str
			 ,char **src_addr,u_int16_t *src_port
			 ,char **if_addr /* for multicast */
			 ,u_int8_t  *ipvar){
	char *strptr[4];
	int  len;
	int  ret;
	strptr[0]=str;
	ret=string_sep(strptr,"/",4,&len);
	switch(len){
	case 4:			/* int  / IP Ver */
		if(*strptr[3]!='\0')
			*ipvar=atoi(strptr[3]);
	case 3:			/* str  / if addr */
		if(*strptr[2]!='\0')
			*if_addr=strptr[2];
	case 2:			/* int  / src port */
		if(*strptr[1]!='\0')
			*src_port=(u_int16_t)atoi(strptr[1]);
	case 1:			/* str  / src addr */
		if(*strptr[0]!='\0')
			*src_addr=strptr[0];
	}
	return(ret);
}


/* 
   MEMO:
   if send packet by multicast. dst_addr is multicast address
   src_addr is a addr this host have.

   str="dst_addr/dst_port/src_addr/src_port/ipvar/ttl"
*/
int string_sep4udpipsend(char *str
			 ,char **dst_addr ,u_int16_t *dst_port
			 ,char **src_addr ,u_int16_t *src_port
			 ,u_int8_t *ipvar
			 ,u_int8_t *ttl){
	char *strptr[6];
	int  len;
	int  ret;
	strptr[0]=str;
	ret=string_sep(strptr,"/",6,&len);
	switch(len){
	case 6:			/* int  / TTL */
		if(*strptr[5]!='\0')
			*ttl=atoi(strptr[5]);
	case 5:			/* int  / IP Ver */
		if(*strptr[4]!='\0')
			*ipvar=atoi(strptr[4]);
	case 4:			/* int  / src port */
		if(*strptr[3]!='\0')
			*src_port=atoi(strptr[3]);
	case 3:			/* str  / src addr */
		if(*strptr[2]!='\0')
			*src_addr=strptr[2];
	case 2:			/* int  / dst port */
		if(*strptr[1]!='\0')
			*dst_port=atoi(strptr[1]);
	case 1:			/* str  / dst addr */
		if(*strptr[0]!='\0')
			*dst_addr=strptr[0];
	}
	return(ret);




}
