/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 *            
 * started:   2002/01/13
 *
 * $Id$
 */

/* 
 * TIME Message Format
 * 
 * 0              7 8            15 16           23 24           31
 * +-------+-+-----+---------------+---------------+---------------+
 * |  Ver  |M| Pad |                 Message Length                | EMON MSG
 * +-------+-+-----+---------------+---------------+---------------+ 
 * |                           Timestamp                           | EMON MSG
 * +---------------+---------------+---------------+---------------+
 * |MSB                        Time (second)                    LSB|
 * +---------------+---------------+---------------+---------------+
 * |MSB                        Time (micro sec)                 LSB|
 * +---------------+---------------+---------------+---------------+
 * 
 * Ver              4-bit Protocol version number = 1.
 * 
 * M                Marker bit = 1.
 * 
 * Messsage Length  24-bit unsigned integer. = 8
 * 
 * Timestamp        32-bit unsigned integer. = 0
 * 
 * Time             Time value 
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>

#include "def_value.h"
#include "pipe_hd.h"
#include "debug.h"
#include "time.h"


#include <sys/types.h>
#include <sys/stat.h>


typedef struct opt_values {	/* set by getenv and/or getopt */
	u_int32_t interval; 
} opt_t;


typedef struct statistics_values{
	int message_count;
} stat_t;


//static stat_t STAT;

int opt_etc(int argc, char *argv[], opt_t *opt);


#define SIZE_OF_TIME
int
main(int argc, char *argv[])
{
	opt_t OPT;
	u_int8_t        p_header[PIPE_HEADER_LEN+sizeof(time_tv_t)];
	u_int8_t       *msg;
	time_tv_t       timeofday;

	if (opt_etc(argc, argv, &OPT) == 0){
		return 1;
	}
	if (isatty(STDOUT)) {
		e_printf("Standard output must be binded with pipe.\n");
		return 1;
	}
	bzero(p_header, PIPE_HEADER_LEN);
	pipe_set_version(p_header, 1);
	pipe_set_marker(p_header, 1);
	pipe_set_length(p_header,sizeof(time_tv_t));
	(u_int8_t*)msg =p_header+PIPE_HEADER_LEN;
	
	e_printf("sizeof(time_tv_t) %d\n",sizeof(time_tv_t) );
	while(1){
		fd_can_write(STDOUT_FILENO,1000);
		timeofday=time_gettimeofday();
		time_htontv(msg,timeofday);
		pipe_blocked_write_block(STDOUT_FILENO
					 ,p_header,msg);
		usleep(OPT.interval);
	};
	return 0;
}


int
opt_etc(int argc, char *argv[], opt_t *opt)
{
	int ch;
	char *opts = "C:R:D:";
	char *helpmes =
		"usage : %s [options]\n"
		" options\n"
		"  -R <n>   : message interval (msec)(default:10000)\n"
		" <debug>\n"
		"  -D <n>   : debug level\n"
		;
	u_int32_t  interval;

	interval = 10000;
	while ((ch = getopt(argc, argv, opts)) != -1){
		switch (ch){
		case 'R':
			interval = atoi(optarg);
			if(interval <=0){
				e_printf("option -r:interval > 0\n");
			}
			break;
		case 'D':
			debug_level = atoi(optarg);
			break;
		default:
			e_printf(helpmes, argv[0]);
			return 0;
		}
	}
	if (argc - optind != 0){
		e_printf(helpmes, argv[0]);
		return 0;
	}
	opt->interval=time_32to32(interval,1000,1000*1000); /* msec=>usec */
	d1_printf("opt:\n"
		  "interval %dmsec\n"
		  ,interval);
	return 1;
}
